"""Config organizer. Provides default config object for velocity."""

from loguru import logger
from platform import processor as arch
from pathlib import Path
from os import getlogin as get_username, getenv
from yaml import safe_load as yaml_safe_load
from ._exceptions import InvalidConfigIdentifier


class Config:
    """Configuration class. Stores configuration as a dictionary."""

    def __init__(self) -> None:
        self._config = dict()

    def set(self, item: str, value: int | bool | str | list | dict | None) -> None:
        """Set configuration property."""
        try:
            # do not let user set the root node
            if item != "":
                parts: list[str] = item.split(":")
                set_value = self._config
                for p in parts:
                    # make all config identifiers comply with python identifiers
                    if not p.isidentifier():
                        raise InvalidConfigIdentifier("'{}' is not a valid identifier.".format(item))
                    # walk config tree until the final node is found
                    if p != parts[-1]:
                        if p not in set_value:
                            set_value[p] = dict()
                        set_value = set_value[p]
                    else:
                        set_value[p] = value
            else:
                raise AttributeError("You cannot set the root config node.")
        except (AttributeError, InvalidConfigIdentifier) as e:
            logger.exception(e)

    def get(self, item: str) -> int | bool | str | list | dict | None:
        """Get configuration property. Return None if not found"""
        try:
            if item != "":
                parts: list[str] = item.split(":")
                ret_value = self._config
                # make all config identifiers comply with python identifiers
                for p in parts:
                    if not p.isidentifier():
                        raise InvalidConfigIdentifier("'{}' is not a valid identifier.".format(item))
                    ret_value = ret_value[p]
                return ret_value
            else:
                return self._config
        except (KeyError, TypeError):
            logger.warning("Could not find '{}' in config.", format(item))
        except InvalidConfigIdentifier as e:
            logger.exception(e)
        return None

    def load(self) -> None:
        """Load configuration."""
        if self.get("velocity:config_dir") is None:
            self.set("velocity:config_dir", Path.home().joinpath(".velocity").__str__())
        config_dir = Path(self.get("velocity:config_dir"))
        try:
            with open(config_dir.joinpath("config.yaml"), "r") as fi:
                conf = yaml_safe_load(fi)
                for k in conf:
                    self.set(k, conf[k])
        except FileNotFoundError:
            logger.warning("Could not load configuration file from '{}'!".format(config_dir.joinpath("config.yaml")))

    def __str__(self) -> str:
        return str(self._config)


# default configuration & singleton
_config = Config()
_config.load()

# get config from environment variables
if getenv("VELOCITY_SYSTEM") is not None:
    _config.set("velocity:system", getenv("VELOCITY_SYSTEM"))

if getenv("VELOCITY_BACKEND") is not None:
    _config.set("velocity:backend", getenv("VELOCITY_BACKEND"))

if getenv("VELOCITY_DISTRO") is not None:
    _config.set("velocity:distro", getenv("VELOCITY_DISTRO"))

if getenv("VELOCITY_CONFIG_DIR") is not None:
    _config.set("velocity:config_dir", getenv("VELOCITY_CONFIG_DIR"))

if getenv("VELOCITY_IMAGE_PATH") is not None:
    _config.set("velocity:image_path", getenv("VELOCITY_IMAGE_PATH"))

if getenv("VELOCITY_BUILD_DIR") is not None:
    _config.set("velocity:build_dir", getenv("VELOCITY_BUILD_DIR"))

# set defaults for un-configured items
if _config.get("velocity:system") is None:
    _config.set("velocity:system", arch())

if _config.get("velocity:backend") is None:
    _config.set("velocity:backend", "apptainer")

if _config.get("velocity:distro") is None:
    _config.set("velocity:distro", "ubuntu")

if _config.get("velocity:debug") is None:
    _config.set("velocity:debug", "WARNING")

if _config.get("velocity:image_path") is None:
    image_dir = Path.home().joinpath(".velocity", "images")
    image_dir.mkdir(parents=True, exist_ok=True)
    _config.set("velocity:image_path", image_dir.__str__())

if _config.get("velocity:build_dir") is None:
    _config.set("velocity:build_dir", Path("/tmp").joinpath(get_username(), "velocity").__str__())

# export
config = _config



"""Exceptions for velocity."""

from sys import stderr


class CannotFindDependency(Exception):
    """Cannot find an image dependency."""

    def __init__(self, *args) -> None:
        super().__init__(*args)


class NoAvailableBuild(Exception):
    """No available build can be found."""

    def __init__(self, *args) -> None:
        super().__init__(*args)


class EdgeViolatesDAG(Exception):
    """Edge breaks the DAG requirement of a graph."""

    def __init__(self, u_of_edge, v_of_edge, cycle) -> None:
        super().__init__(f"Addition of edge {u_of_edge} -> {v_of_edge} violates graph DAG requirement!")
        for c in cycle:
            print(f"{c[0]} -> {c[1]}", file=stderr)


class BackendNotSupported(Exception):
    """Container backend not supported."""

    def __init__(self, *args):
        super().__init__(*args)


class BackendNotAvailable(Exception):
    """Container backend not supported."""

    def __init__(self, *args):
        super().__init__(*args)


class UndefinedVariableInTemplate(Exception):
    """Undefined variable during substitution in template."""

    def __init__(self, *args):
        super().__init__(*args)


class RepeatedSection(Exception):
    """Repeated template section."""

    def __init__(self, *args):
        super().__init__(*args)


class LineOutsideOfSection(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class TemplateSyntaxError(Exception):
    def __init__(self, message, line: str = None):
        super().__init__(f"{message} {f':line: <{line}>' if line is not None else ''}")


class InvalidImageVersionError(Exception):
    """Invalid image version spec."""

    def __init__(self, *args):
        super().__init__(*args)


class InvalidConfigIdentifier(Exception):
    """Invalid configuration identifier."""

    def __init__(self, *args) -> None:
        super().__init__(*args)



from unittest import TestCase
from src.velocity._config import Config

c = Config()
c._config = {
    "test_one": 7,
    "test_two": False,
    "test_three": [5, "a"],
    "test_four": {
        "inner_test": "test"
    }
}

class TestConfig(TestCase):
    def test_set(self):
        c.set("test_four:inner_test", True)
        c.set("one:two:three", "test")
        c.set("again", 56)
        t = {
            "again": 56,
            "test_one": 7,
            "test_two": False,
            "test_three": [5, "a"],
            "test_four": {
                "inner_test": True
            },
            "one": {
                "two": {
                    "three": "test"
                }
            }
        }
        self.assertEqual(t, c._config)

    def test_get(self):
        self.assertEqual(7, c.get("test_one"))
        self.assertFalse(c.get("test_two"))
        self.assertEqual("a", c.get("test_three")[1])
        self.assertEqual("test", c.get("test_four:inner_test"))



from unittest import TestCase
from re import compile
from src.velocity._graph import (
    Version,
    Image
)


class Test(TestCase):
    def test_version_regex(self):
        p = compile(r"^(?P<major>[0-9]+)(?:\.(?P<minor>[0-9]+)(:?\.(?P<patch>[0-9]+))?)?(?:-(?P<suffix>[a-zA-Z0-9]+))?$")

        # test 2.3.5-s90er
        m = p.search("2.3.5-s90er")
        self.assertEqual("2", m.group("major"))
        self.assertEqual("3", m.group("minor"))
        self.assertEqual("5", m.group("patch"))
        self.assertEqual("s90er", m.group("suffix"))

        # test 2.3.5
        m = p.search("2.3.5")
        self.assertEqual("2", m.group("major"))
        self.assertEqual("3", m.group("minor"))
        self.assertEqual("5", m.group("patch"))
        self.assertIsNone(m.group("suffix"))

        # test 2.3
        m = p.search("2.3")
        self.assertEqual("2", m.group("major"))
        self.assertEqual("3", m.group("minor"))
        self.assertIsNone(m.group("patch"))
        self.assertIsNone(m.group("suffix"))

        # test 2
        m = p.search("2")
        self.assertEqual("2", m.group("major"))
        self.assertIsNone(m.group("minor"))
        self.assertIsNone(m.group("patch"))
        self.assertIsNone(m.group("suffix"))


class TestVersion(TestCase):
    def test_init(self):
        # test 2.3.5-s90er
        v = Version("2.3.5-s90er")
        self.assertEqual(2, v.major)
        self.assertEqual(3, v.minor)
        self.assertEqual(5, v.patch)
        self.assertEqual("s90er", v.suffix)

        # test 2.3.5
        v = Version("2.3.5")
        self.assertEqual(2, v.major)
        self.assertEqual(3, v.minor)
        self.assertEqual(5, v.patch)
        self.assertIsNone(v.suffix)

        # test 2.3
        v = Version("2.3")
        self.assertEqual(2, v.major)
        self.assertEqual(3, v.minor)
        self.assertIsNone(v.patch)
        self.assertIsNone(v.suffix)

        # test 2
        v = Version("2")
        self.assertEqual(2, v.major)
        self.assertIsNone(v.minor)
        self.assertIsNone(v.patch)
        self.assertIsNone(v.suffix)

    def test_str(self):
        # test 2.3.5-s90er
        v = Version("2.3.5-s90er")
        self.assertEqual("2.3.5-s90er", v.__str__())

        # test 2.3.5
        v = Version("2.3.5")
        self.assertEqual("2.3.5", v.__str__())

        # test 2.3
        v = Version("2.3")
        self.assertEqual("2.3", v.__str__())

        # test 2
        v = Version("2")
        self.assertEqual("2", v.__str__())

    def test_comparisons(self):
        # test 2.3.4 vs 2.3.4
        one = Version("2.3.4")
        two = Version("2.3.4")
        self.assertTrue(one == two)
        self.assertFalse(one > two)
        self.assertTrue(one >= two)
        self.assertFalse(one < two)
        self.assertTrue(one <= two)

        # test 2.3.4-rc1 vs 2.3.4-rc1
        one = Version("2.3.4-rc1")
        two = Version("2.3.4-rc1")
        self.assertTrue(one == two)
        self.assertFalse(one > two)
        self.assertTrue(one >= two)
        self.assertFalse(one < two)
        self.assertTrue(one <= two)

        # test 2.3.4 vs 2.3.5
        one = Version("2.3.4")
        two = Version("2.3.5")
        self.assertFalse(one == two)
        self.assertFalse(one > two)
        self.assertFalse(one >= two)
        self.assertTrue(one < two)
        self.assertTrue(one <= two)

        # test 2.3.6 vs 2.3.4
        one = Version("2.3.6")
        two = Version("2.3.4")
        self.assertFalse(one == two)
        self.assertTrue(one > two)
        self.assertTrue(one >= two)
        self.assertFalse(one < two)
        self.assertFalse(one <= two)

        # test 2.3.4 vs 2.3.4-rc1
        one = Version("2.3.4")
        two = Version("2.3.4-rc1")
        self.assertTrue(one == two)
        self.assertFalse(one > two)
        self.assertTrue(one >= two)
        self.assertFalse(one < two)
        self.assertTrue(one <= two)

        # test 2.3.4 vs 2.3 & 2
        one = Version("2.3.4")
        two = Version("2.3")
        self.assertTrue(one == two)
        self.assertFalse(one > two)
        self.assertTrue(one >= two)
        self.assertTrue(one <= two)
        self.assertFalse(one < two)
        two = Version("2")
        self.assertTrue(one == two)
        self.assertFalse(one > two)
        self.assertTrue(one >= two)
        self.assertTrue(one <= two)
        self.assertFalse(one < two)


class TestImage(TestCase):
    def test_satisfies_spec(self):
        image = Image("gcc", "12.3.0", "frontier", "apptainer", "ubuntu", "")
        image2 = Image("test", "1", "frontier", "apptainer", "ubuntu", "")
        
        # name
        self.assertTrue(image.satisfies("gcc"))
        self.assertTrue(image2.satisfies("test"))
        self.assertFalse(image.satisfies("ubuntu"))
        
        # version
        self.assertTrue(image.satisfies("gcc@12.3.0"))
        self.assertTrue(image.satisfies("gcc@12.3"))
        self.assertTrue(image.satisfies("gcc@12"))
        self.assertTrue(image.satisfies("gcc@12.3.0:"))
        self.assertTrue(image.satisfies("gcc@:12.3.0"))
        self.assertTrue(image.satisfies("gcc@11.0.0:13.0.0"))
        self.assertTrue(image.satisfies("gcc@:12"))
        self.assertTrue(image.satisfies("gcc gcc@12.3.0"))

        self.assertFalse(image.satisfies("gcc@14.0.0"))
        self.assertFalse(image.satisfies("gcc@1"))
        self.assertFalse(image.satisfies("ubuntu@12.3.0"))
        self.assertFalse(image.satisfies("12.3.0"))
        self.assertFalse(image.satisfies("gcc@13.0.0:13.2.0"))

        # system
        self.assertTrue(image.satisfies("system=frontier"))
        self.assertFalse(image.satisfies("system=summit"))
        self.assertFalse(image.satisfies("frontier"))

        # backend
        self.assertTrue(image.satisfies("backend=apptainer"))
        self.assertFalse(image.satisfies("backend=podman"))
        self.assertFalse(image.satisfies("apptainer"))

        # distro
        self.assertTrue(image.satisfies("distro=ubuntu"))
        self.assertFalse(image.satisfies("dictro=opensuse"))
        self.assertFalse(image.satisfies("ubuntu"))

    #def test_add_spec(self):
    #    self.fail()

    #def test_hash(self):
    #    self.fail()



"""Graph library and tools for dependency graph"""

from loguru import logger
from re import split as re_split, fullmatch as re_fullmatch, Match as ReMatch
from pathlib import Path
from hashlib import sha256
from typing_extensions import Self
from yaml import safe_load as yaml_safe_load
from copy import deepcopy
from enum import Enum
from networkx import (
    DiGraph as nx_DiGraph,
    is_directed_acyclic_graph as nx_is_directed_acyclic_graph,
    find_cycle as nx_find_cycle,
    neighbors as nx_neighbors,
    has_path as nx_has_path,
)
from ._config import config
from ._exceptions import InvalidImageVersionError, CannotFindDependency, EdgeViolatesDAG, NoAvailableBuild


def get_permutations(idx: int, sets: list[list]):
    """
    Get all the possible permutations from a list of lists
    :param idx: recursion level
    :param sets: sets for permutation
    :return: permutations
    """
    permutations = list()
    if len(sets) == 0:
        pass
    elif idx == len(sets) - 1:
        for i in sets[idx]:
            permutations.append({i})
    else:
        for i in sets[idx]:
            sub_permutations = get_permutations(idx + 1, sets)
            for sub in sub_permutations:
                permutations.append(sub.union({i}))
    return permutations


class Version:
    """Version class."""

    def __init__(self, version_specifier: str) -> None:
        self.vs = version_specifier
        try:
            version_dict: dict = re_fullmatch(
                r"^(?P<major>[0-9]+)(?:\.(?P<minor>[0-9]+)(:?\.(?P<patch>[0-9]+))?)?(?:-(?P<suffix>[a-zA-Z0-9]+))?$",
                version_specifier,
            ).groupdict()
            self.major: int | None = int(version_dict["major"]) if version_dict["major"] is not None else None
            self.minor: int | None = int(version_dict["minor"]) if version_dict["minor"] is not None else None
            self.patch: int | None = int(version_dict["patch"]) if version_dict["patch"] is not None else None
            self.suffix: str | None = str(version_dict["suffix"]) if version_dict["suffix"] is not None else None
        except AttributeError:
            self.major: int | None = None
            self.minor: int | None = None
            self.patch: int | None = None
            self.suffix: str | None = None
        except ValueError:
            logger.error("Could not parse version string: {}".format(version_specifier))
            exit(1)

    @property
    def vcs(self) -> str:
        """Version Comparison String"""
        return "{:#>9}.{:#>9}.{:#>9}.{:~<9}".format(
            self.major if self.major is not None else "#",
            self.minor if self.minor is not None else "#",
            self.patch if self.patch is not None else "#",
            self.suffix if self.suffix is not None else "~",
        )

    def preferred(self, other) -> bool:
        """Determine which version to prefer when two version are "equal" ex. 12.3 vs 12.3.0-rc1"""
        return self.vcs > other.vcs

    @property
    def _vcs_t(self) -> str:
        """Version Comparison String Truncated"""
        vl: int = 0
        if self.major is None:
            pass
        elif self.minor is None:
            vl = 9
        elif self.patch is None:
            vl = 19
        elif self.suffix is None:
            vl = 29
        else:
            vl = 39
        return self.vcs[:vl]

    @classmethod
    def _cut_vcses_to_size(cls, one: Self, two: Self) -> tuple[str, str]:
        """Cut vcses to the longest common length"""
        length: int = min(len(one._vcs_t), len(two._vcs_t))
        return one._vcs_t[:length], two._vcs_t[:length]

    def __eq__(self, other) -> bool:
        if not isinstance(other, Version):
            raise TypeError(
                f"'>' not supported between instances of " f"'{type(self).__name__}' and '{type(other).__name__}'"
            )
        s_vcs, o_vcs = self._cut_vcses_to_size(self, other)
        return s_vcs == o_vcs

    def __ne__(self, other) -> bool:
        if not isinstance(other, Version):
            raise TypeError(
                f"'>' not supported between instances of " f"'{type(self).__name__}' and '{type(other).__name__}'"
            )
        s_vcs, o_vcs = self._cut_vcses_to_size(self, other)
        return s_vcs != o_vcs

    def __gt__(self, other) -> bool:
        if not isinstance(other, Version):
            raise TypeError(
                f"'>' not supported between instances of " f"'{type(self).__name__}' and '{type(other).__name__}'"
            )
        return self.vcs > other.vcs and not self.__eq__(other)

    def __ge__(self, other) -> bool:
        if not isinstance(other, Version):
            raise TypeError(
                f"'>' not supported between instances of " f"'{type(self).__name__}' and '{type(other).__name__}'"
            )
        return self.vcs > other.vcs or self.__eq__(other)

    def __lt__(self, other) -> bool:
        if not isinstance(other, Version):
            raise TypeError(
                f"'>' not supported between instances of " f"'{type(self).__name__}' and '{type(other).__name__}'"
            )
        return self.vcs < other.vcs and not self.__eq__(other)

    def __le__(self, other) -> bool:
        if not isinstance(other, Version):
            raise TypeError(
                f"'>' not supported between instances of " f"'{type(self).__name__}' and '{type(other).__name__}'"
            )
        return self.vcs < other.vcs or self.__eq__(other)

    def __str__(self) -> str:
        return self.vs


class Image:
    """Velocity container image."""

    def __init__(self, name: str, version: str, system: str, backend: str, distro: str, path: str) -> None:
        # foundational
        self.name: str = name
        self.version: Version = Version(str(version))
        self.system: str = system
        self.backend: str = backend
        self.distro: str = distro
        self.dependencies: set[str] = set()

        # additional
        self.variables: dict[str, str] = dict()
        self.arguments: set[str] = set()
        self.template: str = "default"
        self.files: set[str] = set()
        self.prolog: str | None = None
        self.underlay: int | None = None  # sum of the ids this image will be built on

        # metadata
        self.path: Path = Path(path)

    def satisfies(self, spec: str) -> bool:
        """Test if this node satisfies the given spec."""
        name_version_regex: str = (
            r"^(?P<name>{})(?:(?:@(?P<left>[\d\.]+)(?!@))?(?:@?(?P<colen>:)(?P<right>[\d\.]+)?)?)?$".format(self.name)
        )
        system_regex: str = r"^system=(?P<system>[a-zA-Z0-9]+)$"
        backend_regex: str = r"^backend=(?P<backend>[a-zA-Z0-9]+)$"
        distro_regex: str = r"^distro=(?P<distro>[a-zA-Z0-9]+)$"
        dependency_regex: str = r"^\^(?P<name>[a-zA-Z0-9-]+)$"

        ss: list[str] = re_split(r"\s+", spec.strip())

        for part in ss:
            # name and version
            res: ReMatch | None = re_fullmatch(name_version_regex, part)
            if res is not None:
                gd: dict = res.groupdict()
                if gd["left"] is not None and gd["right"] is None:  # n@v: or n@v
                    if gd["colen"] is not None:
                        if Version(gd["left"]) > self.version:
                            return False
                    else:
                        if Version(gd["left"]) != self.version:
                            return False
                elif gd["left"] is None and gd["right"] is not None:  # n@:v
                    if gd["colen"] is not None:
                        if Version(gd["right"]) < self.version:
                            return False
                    else:
                        return False
                elif gd["left"] is None and gd["right"] is None:  # n
                    if gd["colen"] is not None:
                        return False
                else:  # n@v:v
                    if Version(gd["left"]) > self.version or self.version > Version(gd["right"]):
                        return False
                continue  # part has been handled so continue

            # system
            res = re_fullmatch(system_regex, part)
            if res is not None:
                if res.group("system") != self.system:
                    return False
                continue  # part has been handled so continue

            # backend
            res = re_fullmatch(backend_regex, part)
            if res is not None:
                if res.group("backend") != self.backend:
                    return False
                continue  # part has been handled so continue

            # distro
            res = re_fullmatch(distro_regex, part)
            if res is not None:
                if res.group("distro") != self.distro:
                    return False
                continue  # part has been handled so continue

            # dependencies
            res = re_fullmatch(dependency_regex, part)
            if res is not None:
                matched = False
                for dep in self.dependencies:
                    if res.group("name") == dep:
                        matched = True
                if matched:
                    continue  # match is found so continue

            # if we get here this part has not been handled
            return False

        # all parts were handled
        return True

    def apply_constraint(self, conditional: str, _type: str, spec: str) -> bool:
        """Evaluate and apply constraints. Return True if a constraint changes the dependencies."""
        if self.satisfies(conditional):
            if _type == "dependency":
                if spec not in self.dependencies:
                    self.dependencies.add(spec)
                    return True
            elif _type == "variable":
                parts = spec.split("=")
                self.variables[parts[0]] = parts[1]
            elif _type == "argument":
                self.arguments.add(spec)
            elif _type == "template":
                self.template = spec
            elif _type == "file":
                self.files.add(spec)
            elif _type == "prolog":
                self.prolog = spec
        return False

    @property
    def hash(self) -> str:
        """Return a hash for this node uniquely identifying it."""

        hash_list: list = list()
        hash_list.append(self.name)
        hash_list.append(self.version)
        hash_list.append(self.system)
        # hash_list.append(self.backend) # disable backend for now because it should not make a difference in the image
        hash_list.append(self.distro)
        hash_list.append(",".join(str(x) for x in self.dependencies))
        hash_list.append(",".join(str(x) for x in self.variables))
        hash_list.append(",".join(str(x) for x in self.arguments))
        tf = Path(self.path).joinpath("templates", "{}.vtmp".format(self.template))
        if tf.is_file():
            hash_list.append(sha256(tf.read_bytes()).hexdigest())
        else:
            hash_list.append(None)
        hash_list.append(",".join(str(x) for x in self.files))
        hash_list.append(self.prolog)
        hash_list.append(self.underlay)

        hash_str: str = "|".join(str(x) for x in hash_list)
        return sha256(hash_str.encode()).hexdigest()

    @property
    def id(self) -> str:
        """Short hash."""
        return self.hash[:7]

    def __hash__(self) -> int:
        return int(self.hash, 16)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Image):
            return False
        return self.hash == other.hash

    def __lt__(self, other) -> bool:
        if isinstance(other, Image):
            return not self.version.preferred(other.version)
        return False

    def __str__(self) -> str:
        return "{} {}@{} system={} backend={} distro={}{}".format(
            self.hash[:7],
            self.name,
            self.version,
            self.system,
            self.backend,
            self.distro,
            "".join(" ^{}".format(x) for x in self.dependencies) if len(self.dependencies) > 0 else "",
        )


class DepOp(Enum):
    """Dependency options."""

    EQ = "="
    GE = ">="
    LE = "<="
    UN = None


class Target:
    """Build targets."""

    def __init__(self, node: Image, op: DepOp):
        self.node = node
        self.op = op

    def __str__(self):
        return "Target: {} -> {}".format(self.op, self.node)


class ImageGraph(nx_DiGraph):
    """Image dependency graph."""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def add_edge(self, u_of_edge: Image, v_of_edge: Image, **kwargs) -> None:
        # check that edge endpoints are in graph
        if self.has_node(u_of_edge) and self.has_node(v_of_edge):
            super().add_edge(u_of_edge, v_of_edge, **kwargs)
        else:
            raise CannotFindDependency("Cannot find dependency {} for {}".format(v_of_edge, u_of_edge))

        # check that graph is still a DAG
        if not nx_is_directed_acyclic_graph(self):
            cycle = nx_find_cycle(self)
            raise EdgeViolatesDAG(u_of_edge, v_of_edge, cycle)

    def get_similar_nodes(self, node: Image) -> set:
        """Get all nodes with the same name."""
        similar = set()
        for n in self.nodes:
            if n.satisfies(node.name):
                similar.add(n)
        return similar

    def get_dependencies(self, node: Image) -> set[Image]:
        """Get all dependencies for an image."""
        # had to add this wrapper because nx.neighbors was dropping the attributes of some nodes (cuda, python)
        deps = set()
        for node in nx_neighbors(self, node):
            for n in self.nodes:
                if n == node:
                    deps.add(n)
        # TODO will contact the maintainers of networkx at some point
        return deps

    def is_above(self, u_node: Image, v_node: Image) -> bool:
        """Test if one node is above another in the dependency tree."""
        return nx_has_path(self, u_node, v_node)

    def _is_valid_build_tuple(self, bt: tuple[Image]) -> bool:
        """Verify that all the dependencies of a build tuple can be met."""
        valid = True

        # check for similar images
        for i0 in range(len(bt)):
            for i2 in bt[i0 + 1 :]:
                if bt[i0].satisfies(i2.name):
                    valid = False

        # check that all images exist
        for node in bt:
            if not self.has_node(node):
                valid = False

        # break prematurely if the first two checks fail
        if not valid:
            return valid

        # check that deps in build tuple
        for node in bt:
            deps = self.get_dependencies(node)

            # group deps
            grouped = dict()
            for d in deps:
                if d.name not in grouped:
                    grouped[d.name] = set()
                grouped[d.name].add(d)

            # check that the needed dependency exists
            for g in grouped:
                if grouped[g].isdisjoint(bt):
                    valid = False

        return valid

    def create_build_recipe(self, targets: list[Target]) -> tuple:
        """Create a build recipe."""
        # check if all the targets exist
        for node in targets:
            if len(self.get_similar_nodes(node.node)) < 1:
                raise NoAvailableBuild(f"The build target {node.node} does not exist!")

        # init build set and priority list
        build_set = set()
        priority_list = list()

        # add similar to build set
        for target in targets:
            build_set.update(self.get_similar_nodes(target.node))
            # make sure to update priority
            if target.node.name not in priority_list:
                priority_list.append(target.node.name)

        # add deps to build set
        while True:
            build_set_length = len(build_set)

            for node in build_set.copy():
                dependencies = self.get_dependencies(node)
                for d in dependencies:
                    build_set.add(d)
                    if d.name not in priority_list:
                        priority_list.append(d.name)

            # loop until all dependencies are added
            if build_set_length == len(build_set):
                break

        # apply constraints
        for target in targets:
            for node in build_set.copy():
                if node.satisfies(target.node.name):
                    if target.op == DepOp.EQ and node.version != target.node.version:
                        build_set.remove(node)
                    elif target.op == DepOp.GE and node.version < target.node.version:
                        build_set.remove(node)
                    elif target.op == DepOp.LE and node.version > target.node.version:
                        build_set.remove(node)

        # group deps
        grouped = dict()
        for node in build_set:
            if node.name not in grouped:
                grouped[node.name] = set()
            grouped[node.name].add(node)
        # sort deps so that the highest versions of images further up the dep tree will be chosen
        prioritized_list_group = list()
        for group in priority_list:
            tmp = list(grouped[group])
            tmp.sort(reverse=True)
            prioritized_list_group.append(tmp)
        # get permutations
        permutations = get_permutations(0, prioritized_list_group)

        # return valid build tuple
        for p in permutations:
            # clean up permutation
            clean_p = set()
            for n in p:
                for t in targets:
                    con_targ = None
                    for nt in [x for x in p]:
                        if nt.satisfies(t.node.name):
                            con_targ = nt
                            break
                    if self.is_above(con_targ, n):
                        clean_p.add(n)
                        break

            if self._is_valid_build_tuple(tuple(clean_p)):
                # order build
                build_list = list()
                processed = set()
                unprocessed = clean_p.copy()
                while len(unprocessed) > 0:
                    level_holder = list()
                    for node in unprocessed.copy():
                        deps = set(self.get_dependencies(node)).intersection(clean_p)
                        if deps.issubset(processed):
                            level_holder.append(node)
                    level_holder.sort()
                    for node in level_holder:
                        processed.add(node)
                        unprocessed.remove(node)
                        build_list.append(node)

                return tuple(build_list)

        # if we got here no valid build tuple could be found
        raise NoAvailableBuild("No Available build!")


class ImageRepo:
    """Image repository."""

    def __init__(self) -> None:
        self.images: set[Image] = set()
        # constraint(image, condition, type, spec, scope)
        self.constraints: list[tuple[str, str, str, str, str]] = list()
        if config.get("constraints") is not None:
            # arguments
            if "arguments" in config.get("constraints"):
                for argument in config.get("constraints")["arguments"]:
                    if isinstance(argument["value"], list):
                        specs = argument["value"]
                    else:
                        specs = [
                            argument["value"],
                        ]
                    for spec in specs:
                        self.constraints.append(
                            (
                                "",
                                argument["when"] if "when" in argument else "",
                                "argument",
                                spec,
                                "global",
                            )
                        )
            # variables
            if "variables" in config.get("constraints"):
                for variable in config.get("constraints")["variables"]:
                    self.constraints.append(
                        (
                            "",
                            variable["when"] if "when" in variable else "",
                            "variable",
                            "{}={}".format(variable["name"], variable["value"]),
                            "global",
                        )
                    )

    def import_from_dir(self, path: str) -> None:
        """Add Images from path."""
        p = Path(path)
        if not p.is_dir():
            raise NotADirectoryError(f"Image path {path} is not a directory!")

        for name in [x for x in p.iterdir() if x.is_dir() and x.name[0] != "."]:
            with open(name.joinpath("specs.yaml"), "r") as fi:
                specs_file = yaml_safe_load(fi)
                # add versions
                for version in specs_file["versions"]:
                    if isinstance(version["spec"], list):
                        specs = version["spec"]
                    else:
                        specs = [
                            version["spec"],
                        ]
                    for spec in specs:
                        image = Image(
                            name.name,
                            spec,
                            config.get("velocity:system"),
                            config.get("velocity:backend"),
                            config.get("velocity:distro"),
                            str(name),
                        )
                        if "when" in version:
                            if image.satisfies(version["when"]):
                                self.images.add(image)
                        else:
                            self.images.add(image)
                # add constraints
                # dependencies
                if "dependencies" in specs_file:
                    for dependency in specs_file["dependencies"]:
                        if isinstance(dependency["spec"], list):
                            specs = dependency["spec"]
                        else:
                            specs = [
                                dependency["spec"],
                            ]
                        for spec in specs:
                            self.constraints.append(
                                (
                                    name.name,
                                    dependency["when"] if "when" in dependency else "",
                                    "dependency",
                                    spec,
                                    dependency["scope"] if "scope" in dependency else "image",
                                )
                            )
                # templates
                if "templates" in specs_file:
                    for template in specs_file["templates"]:
                        if isinstance(template["name"], list):
                            specs = template["name"]
                        else:
                            specs = [
                                template["name"],
                            ]
                        for spec in specs:
                            self.constraints.append(
                                (
                                    name.name,
                                    template["when"] if "when" in template else "",
                                    "template",
                                    spec,
                                    template["scope"] if "scope" in template else "image",
                                )
                            )
                # arguments
                if "arguments" in specs_file:
                    for argument in specs_file["arguments"]:
                        if isinstance(argument["value"], list):
                            specs = argument["value"]
                        else:
                            specs = [
                                argument["value"],
                            ]
                        for spec in specs:
                            self.constraints.append(
                                (
                                    name.name,
                                    argument["when"] if "when" in argument else "",
                                    "argument",
                                    spec,
                                    argument["scope"] if "scope" in argument else "image",
                                )
                            )
                # variables
                if "variables" in specs_file:
                    for variable in specs_file["variables"]:
                        self.constraints.append(
                            (
                                name.name,
                                variable["when"] if "when" in variable else "",
                                "variable",
                                "{}={}".format(variable["name"], variable["value"]),
                                variable["scope"] if "scope" in variable else "image",
                            )
                        )
                # files
                if "files" in specs_file:
                    for file in specs_file["files"]:
                        if isinstance(file["name"], list):
                            specs = file["name"]
                        else:
                            specs = [
                                file["name"],
                            ]
                        for spec in specs:
                            self.constraints.append(
                                (
                                    name.name,
                                    file["when"] if "when" in file else "",
                                    "file",
                                    spec,
                                    file["scope"] if "scope" in file else "image",
                                )
                            )
                # prologs
                if "prologs" in specs_file:
                    for prolog in specs_file["prologs"]:
                        self.constraints.append(
                            (
                                name.name,
                                prolog["when"] if "when" in prolog else "",
                                "prolog",
                                prolog["script"],
                                prolog["scope"] if "scope" in prolog else "image",
                            )
                        )

    def create_build_recipe(self, targets: list[str]) -> tuple[tuple, ImageGraph]:
        """Create an ordered build recipe of images."""
        images: set[Image] = deepcopy(self.images)

        build_targets: list[Target] = list()
        for target in targets:
            res = re_fullmatch(
                r"^(?P<name>[a-zA-Z0-9-]+)(?:(?:@(?P<left>[\d\.]+)(?!@))?(?:@?(?P<colen>:)(?P<right>[\d\.]+)?)?)?$",
                target,
            )
            if res is not None:
                gd: dict = res.groupdict()
                # n@v: or n@v
                if gd["left"] is not None and gd["right"] is None:
                    t = Image(
                        gd["name"],
                        gd["left"],
                        config.get("velocity:system"),
                        config.get("velocity:backend"),
                        config.get("velocity:distro"),
                        "",
                    )
                    if gd["colen"] is not None:
                        build_targets.append(Target(t, DepOp.GE))
                    else:
                        build_targets.append(Target(t, DepOp.EQ))
                # n@:v
                elif gd["left"] is None and gd["right"] is not None:
                    if gd["colen"] is not None:
                        t = Image(
                            gd["name"],
                            gd["right"],
                            config.get("velocity:system"),
                            config.get("velocity:backend"),
                            config.get("velocity:distro"),
                            "",
                        )
                        build_targets.append(Target(t, DepOp.LE))
                    else:
                        raise InvalidImageVersionError("Invalid version '{}'.".format(target))
                # n
                elif gd["left"] is None and gd["right"] is None:
                    if gd["colen"] is not None:
                        raise InvalidImageVersionError("Invalid version '{}'.".format(target))
                    else:
                        t = Image(
                            gd["name"],
                            "",
                            config.get("velocity:system"),
                            config.get("velocity:backend"),
                            config.get("velocity:distro"),
                            "",
                        )
                        build_targets.append(Target(t, DepOp.UN))
                # n@v:v
                else:
                    t = Image(
                        gd["name"],
                        gd["left"],
                        config.get("velocity:system"),
                        config.get("velocity:backend"),
                        config.get("velocity:distro"),
                        "",
                    )
                    build_targets.append(Target(t, DepOp.GE))
                    t = Image(
                        gd["name"],
                        gd["right"],
                        config.get("velocity:system"),
                        config.get("velocity:backend"),
                        config.get("velocity:distro"),
                        "",
                    )
                    build_targets.append(Target(t, DepOp.LE))
            else:
                raise NoAvailableBuild("No available build!")

        # pre-burner graph
        for constraint in self.constraints:
            pass
            for image in images:
                if image.apply_constraint("{} {}".format(constraint[0], constraint[1]), constraint[2], constraint[3]):
                    images_changed = True
        ig = ImageGraph()
        for image in images:
            ig.add_node(image)
        for image in images:
            for dep in image.dependencies:
                for di in images:
                    if di.satisfies(dep):
                        ig.add_edge(image, di)

        bt: tuple[Image] = ig.create_build_recipe(build_targets)

        # apply constraints for the build scope
        images_changed: bool = True
        while images_changed:
            images_changed = False
            for constraint in self.constraints:
                if constraint[4] == "build":
                    for targ in bt:
                        if targ.satisfies(constraint[1]):
                            for image in images:
                                if image.apply_constraint(constraint[0], constraint[2], constraint[3]):
                                    images_changed = True
                # "image" or "universal"
                else:
                    for image in images:
                        if image.apply_constraint(
                            "{} {}".format(constraint[0], constraint[1]), constraint[2], constraint[3]
                        ):
                            images_changed = True

        # create graph
        ig = ImageGraph()
        for image in images:
            ig.add_node(image)
        for image in images:
            for dep in image.dependencies:
                for di in images:
                    if di.satisfies(dep):
                        ig.add_edge(image, di)

        bt: tuple[Image] = ig.create_build_recipe(build_targets)

        # update images so that their hash includes the layers below them
        cumulative_deps: int = 0
        for b in bt:
            b.underlay = cumulative_deps
            cumulative_deps = cumulative_deps + int(b.id, 16)

        bt_ig = ImageGraph()
        for image in bt:
            bt_ig.add_node(image)
        for image in bt:
            for dep in image.dependencies:
                for di in bt:
                    if di.satisfies(dep):
                        bt_ig.add_edge(image, di)

        return bt, bt_ig



"""Run velocity as a script."""

import argparse
import sys
from loguru import logger
from importlib.metadata import version
from colorama import Fore, Style
from ._graph import ImageRepo, Image
from ._build import ImageBuilder
from ._print import header_print, indent_print, TextBlock, bare_print
from ._config import config

############################################################
# Parse Args
############################################################
parser = argparse.ArgumentParser(
    prog="velocity",
    description="build tool for OLCF containers",
    epilog="See https://github.com/olcf/velocity",
)
parser.add_argument(
    "-v", "--version", action="version", version=f"%(prog)s {version('olcf-velocity')}", help="program version"
)
parser.add_argument(
    "-D",
    "--debug",
    choices=["TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"],
    help="set debug output level",
)
parser.add_argument("-b", "--backend")
parser.add_argument("-s", "--system")
parser.add_argument("-d", "--distro")

# create sub_parsers
sub_parsers = parser.add_subparsers(dest="subcommand")

# create build_parser
build_parser = sub_parsers.add_parser("build", help="build specified container image")
build_parser.add_argument("-d", "--dry-run", action="store_true", help="dry run build system")
build_parser.add_argument("targets", type=str, nargs="+", help="build targets")
build_parser.add_argument("-n", "--name", action="store", help="name of complete image")
build_parser.add_argument(
    "-l", "--leave-tags", action="store_true", help="do not clean up intermediate build tags (only applies to podman)"
)
build_parser.add_argument("-v", "--verbose", action="store_true", help="print helpful debug/runtime information")
build_parser.add_argument("-c", "--clean", action="store_true", help="run clean build (delete cached builds)")

# create avail_parser
avail_parser = sub_parsers.add_parser("avail", help="lookup available images")

# create spec_parser
spec_parser = sub_parsers.add_parser("spec", help="lookup image dependencies")
spec_parser.add_argument("targets", type=str, nargs="+", help="spec targets")

# parse args
args = parser.parse_args()

############################################################
# apply user run time arguments over settings
############################################################
if args.debug is not None:
    config.set("velocity:debug", args.debug)
if args.system is not None:
    config.set("velocity:system", args.system)
if args.backend is not None:
    config.set("velocity:backend", args.backend)
if args.distro is not None:
    config.set("velocity:distro", args.distro)

# setup logging and log startup
logger.enable("velocity")
logger.configure(handlers=[{"sink": sys.stdout, "level": config.get("velocity:debug")}])
logger.debug(
    "Starting velocity {},{},{}.".format(
        config.get("velocity:system"), config.get("velocity:backend"), config.get("velocity:distro")
    )
)
logger.trace(config.get(""))

############################################################
# Load images
############################################################
imageRepo = ImageRepo()
for p in config.get("velocity:image_path").split(":"):
    imageRepo.import_from_dir(p)

############################################################
# Handle User Commands
############################################################
if args.subcommand == "build":
    # get recipe
    recipe = imageRepo.create_build_recipe(args.targets)[0]

    # print build specs
    header_print([TextBlock("Build Order:")])
    for r in recipe:
        indent_print([TextBlock(f"{r.name}@{r.version}-{r.id}", fore=Fore.MAGENTA, style=Style.BRIGHT)])
    print()  # newline

    # prep builder
    builder = ImageBuilder(
        recipe,
        build_name=args.name,
        dry_run=args.dry_run,
        remove_tags=not args.leave_tags,
        verbose=args.verbose,
        clean_build_dir=args.clean
    )

    # build
    builder.build()

elif args.subcommand == "avail":
    # group and order
    grouped = dict()
    for node in imageRepo.images:
        if node.name not in grouped:
            grouped[node.name] = list()
        grouped[node.name].append(node)
    ordered = list(grouped.keys())
    ordered.sort()

    # print
    for group in ordered:
        header_print([TextBlock(group, fore=Fore.RED, style=Style.BRIGHT)])
        deps = grouped[group]
        deps.sort()
        for t in deps:
            indent_print([TextBlock(t.version, fore=Fore.YELLOW, style=Style.BRIGHT)])
    print()  # add newline

elif args.subcommand == "spec":
    # get recipe
    recipe, graph = imageRepo.create_build_recipe(args.targets)

    # flatten dependency tree
    flat_dep_tree = dict()
    for r in recipe:
        flat_dep_tree[r.name] = set()
        deps = set(graph.get_dependencies(r))
        for o in deps.intersection(set(recipe)):
            flat_dep_tree[r.name].add(o.name)
    # get top level entries
    top_level_entries = set()
    deps = set()
    for r in recipe:
        deps.update(graph.get_dependencies(r))
    for r in recipe:
        if r not in deps:
            top_level_entries.add(r)

    def spec_print(seed: str, indent: int, fdt: dict, rs: tuple[Image]):
        """
        Recursive function to print dep tree
        """
        spec = ""
        for _ in rs:
            if _.satisfies(seed):
                spec = "{}@{}-{}".format(_.name, _.version, _.id)
        print("  " + "   " * indent, end="")
        if indent == 0:
            bare_print(
                [
                    TextBlock("> ", fore=Fore.RED, style=Style.BRIGHT),
                    TextBlock(spec, fore=Fore.MAGENTA, style=Style.BRIGHT),
                ]
            )
        else:
            bare_print(
                [
                    TextBlock("^", fore=Fore.GREEN, style=Style.BRIGHT),
                    TextBlock(spec, fore=Fore.MAGENTA, style=Style.BRIGHT),
                ]
            )
        if len(fdt[seed]) > 0:
            for un in fdt[seed]:
                spec_print(un, (indent + 1), fdt, rs)

    # print specs
    for tl in top_level_entries:
        spec_print(tl.name, 0, flat_dep_tree, recipe)
    print()  # add newline
else:
    parser.print_help()
    print()  # add newline
